
(* This file describes the instructions and the memory of the VM *)

(* binary operations, taking the accumulator and the top of the stack
   and putting their result back in the accumulator *)
type binop =
  | Add
  | Sub
  | Mul
  | Div
  | Eqi
  | Cat

(* the instructions of the VM parameterized by the representation of
   labels (see the ['label] type parameter). *)
type 'label instruction =
  | Halt
  | Binop of binop
  | Const of int
  | Str of string
  | Push
  | Acc of int
  | Print
  | Apply
  | Return of int
  | Pop of int
  | Makeblock of int * int
  | Getblock of int
  | Closure of int * 'label
  | Branchif of 'label
  | Branch of 'label

(* the intermediate code is a list of instruction where labels are
   themselves intermediate code. In intermediate code:
   - a [Branch c] only indicates a *block* of code.
   - if the accumulator is different from 0, [Branchif c] execute [c] and then
     jump out of the directly enclosing block
   - in [Closure (n, c)], the c is the code of the function, with the [Return]
     instruction implicit.
*)
type interm = Seq of interm instruction list

(* in real bytecode, labels are represented by offsets, that is the
   number of instructions to jump forward *)
type offset = int
type bytecode = offset instruction array

type tag = int

(* a block can hold either a string or an array of words with a tag *)
type block =
  | String of string
  | Block of tag * word array

(* a memory word is either a direct value, represented as an integer,
   or a pointer to a block *)
and word =
  | Int of int
  | Ptr of block

(* the state of the VM consists in some code, a pc, an accumulator and
   a stack. All memory cells can hold a word. *)
type state = {
  code : bytecode;
  pc : int;
  acc : word;
  stack : word Stk.t;
}
